/********************************************************************
*      IMPLEMENTACAO DE UMA ULA/UC SEGUNDO A ARQUITETURA RISC       *
* ----------------------------------------------------------------- *
*      Membros: Anderson de Rezende Rocha - DCC/UFLA                *
*               undersun@comp.ufla.br                              *
*                                                                   *
*               Jlio Csar Alves - DCC/UFLA                        *
*               jcalves@comp.ufla.br                                *
* ----------------------------------------------------------------- *
*      ARQUITETURA DE COMPUTADORES. Prof. Mrio Luis R. Oliveira    *
********************************************************************/

/*  OBSERVAES DE FUNCIONAMENTO
    * No h suporte a nmeros negativos
    * Os dados no arquivo de entrada devem todos estar em binrio
    * Ao se referir a um registrador use 4 bits como 0000
    * Ao se referir a um identificador de instrucao use 4 bits como 0000
    * Ao se referir a uma posicao de memoria use 8 bits como 00001000


    INSTRUES DISPONVEIS
    + 0000 NOP                     % Instruo de vcuo
    + 0001 ADD    R1 R2 R3         % R1 = R2 + R3
    + 0010 SUB    R1 R2 R3         % R1 = R2 - R3
    + 0011 MULT   R1 R2 R3         % R1 = R2 * R3
    + 0100 DIV    R1 R2 R3         % R1 = R2 / R3
    + 0101 AND    R1 R2 R3         % R1 = R2 & R3
    + 0110 OR     R1 R2 R3         % R1 = R2 | R3
    + 0111 XOR    R1 R2 R3         % R1 = R2 ^ R3
    + 1000 NOT    R1 R2            % R1 = ~R2
    + 1001 CPOP   R1 R2            % R1 = R2
    + 1010 LOAD   R1 MEM           % R1 = MEM
    + 1011 STORE  R1 MEM           % MEM = R1
    + 1100 SHIFTL R1 R2 R3         % R1 = R2 << R3
    + 1101 SHIFTR R1 R2 R3         % R1 = R2 >> R3
    + 1110 MOD    R1 R2 R3         % R1 = R2 % R3
    + 1111 HALT                    % FINALIZA A EXECUCAO

    MTODOS DISPONVEIS
    + Risc();
    + void Set(string mapaMemoria, string mapaInstrucoes);
    + void ImprimirStatus();
    + void ProcessarPassoAPasso();
    + void ProcessarTudo();

    - string ToBit(unsigned n, bool complemento = false);
    - unsigned ToInt(string s);
    - bool MapearInstrucoes(string arq);
    - bool MapearMemoria(string arq);
    - bool emMemoria;
    - void BuscarInstrucao();
    - void AtualizarPC();
    - void Decodificar();
    - void VerificarOperandos();
    - void BuscarOperandos();
    - void EscreverResultado();
    - void Executar();
*/

#ifndef RISC_H
#define RISC_H

#include <iostream>
#include <fstream>
#include <string>
#include <iomanip>
#include <math.h>

using namespace std;

// Todas as instrucoes disponiveis no sistema
enum E_INSTRUCOES {NOP = 0, ADD, SUB, MULT, DIV, AND, OR, XOR, NOT,
                CPOP, LOAD, STORE, SHIFTL, SHIFTR, MOD, HALT, };
class Risc
{
    public:
        // Construtor padrao
        Risc();
        // Seta os arquivos para mapeamento de memoria e instrucoes
        // Caso queira executar outro conjunto de instrucoes basta
        // chamar este metodo
        void Set(string mapaMemoria, string mapaInstrucoes);
        void ImprimirStatus();
        bool ProcessarPassoAPasso();
        void ProcessarTudo();
        
        // prepara o simulador para ser reiniciado
        void Limpar();

        string GetMemoria(int i)
        {
            return memoria[i];
        }

        string GetRegistradores(int i)
        {
            return registradores[i];
        }

        int GetCiclosClock()
        {
            return ciclosClock;
        }

        int GetCache(int i)
        {
            return cache[i];
        }

        // A partir do arquivo passado com as instrucoes passa as mesmas para
        // a memoria do simulador
        bool MapearInstrucoes(string arq);

        // A partir do arquivo passado com o mapa de memoria passa o mesmo para
        // a memoria do simulador
        bool MapearMemoria(string arq);


    private:
        // O usuario e totalmente responsavel pelo uso da memoria
        // No  verificado a sobreposicao de valores na memoria nem
        // se algum valor esta substituindo uma palavra do codigo de
        // instrucoes
        string memoria[256];

        // R0: Contador de programa [PC]
        // R1: Registrador de instrucao
        // [R2-R15] RPGs
        string registradores[16];

        // Contagem geral de quantos ciclos de clock um determinado
        // programa utilizou;
        int ciclosClock;

        // Cache do processador para guardar o resultado de uma operacao
        // c[0] opcod
        // c[1] op1
        // c[2] op2
        // c[3] op3
        // c[4] resultado de algum calculo
        int cache[5];

        // Transforma um numero n qualquer em sua sequencia de bits
        string ToBit(unsigned n, bool complemento = false);

        // Transforma uma sequencia binaria representada na forma de string
        // como em "0010" em seu correspondente decimal, no caso, 2.
        unsigned ToInt(string s);

        // Identifica se alguma instrucao deve acessar a memoria
        bool emMemoria;

        // As SETE tradicionais instrues de um sistema RISC
        // Todas as instrues so executadas em 7 ciclos
        // CICLO 00
        void BuscarInstrucao();
        // CICLO 01
        void AtualizarPC();
        // CICLO 02
        void Decodificar();
        // CICLO 03
        void VerificarOperandos();
        // CICLO 04
        void BuscarOperandos();
        // CICLO 05
        void Executar();
        // CICLO 06
        void EscreverResultado();        
};

// Seta os arquivos para mapeamento de memoria e instrucoes
// Caso queira executar outro conjunto de instrucoes basta
// chamar este metodo
void Risc::Set(string mapaMemoria, string mapaInstrucoes)
{
    MapearMemoria(mapaMemoria);
    MapearInstrucoes(mapaInstrucoes);
}

// Construtor padrao
Risc::Risc()
{
    Limpar();
}

void Risc::Limpar()
{
    emMemoria = false;
    ciclosClock = 0;
    // Contador de programas e registrador de instrues precisam
    // ser inicializados
    registradores[0] = "0000000000000000";
    registradores[1] = "0000000000000000";
    for (int i = 2; i < 16; i++)
        registradores[i] = "";
    
    for (int i = 0; i < 256; i++)
        memoria[i] = "";
        
    cache[0] = cache[1] = cache[2] = cache[3] = cache[4] = 0;    
}

// Carrega as instrues do sistema de um arquivo para a memria
// do simulador
bool Risc::MapearInstrucoes(string arq)
{
    ifstream entrada(arq.c_str());
    string opcod, op1, op2, op3;
    string palavra;
    int valorOpcod, pos = 0;

    while(!entrada.eof())
    {
        opcod = op1 = op2 = op3 = "";
        entrada >> opcod;
        valorOpcod = ToInt(opcod);

        switch (valorOpcod)
        {
            case NOP:
                palavra = ToBit(0);
                break;
            case ADD:
            case SUB:
            case MULT:
            case DIV:
            case MOD:
            case AND:
            case OR:
            case XOR:
            case SHIFTL:
            case SHIFTR:
                entrada >> op1;
                entrada >> op2;
                entrada >> op3;
                palavra = opcod;
                palavra += op1;
                palavra += op2;
                palavra += op3;
                break;
            case NOT:
            case CPOP:
                entrada >> op1;
                entrada >> op2;
                palavra = opcod;
                palavra += op1;
                palavra += op2;
                palavra += "0000";
                break;
            case LOAD:
            case STORE:
                entrada >> op1;
                entrada >> op2;
                palavra = opcod;
                palavra += op1;
                palavra += op2;
                break;
            case HALT:
                palavra = opcod;
                palavra += "000000000000";
                break;
            default:
                cerr << "Instrucao desconhecida" << endl;
                return false;
        }
        memoria[pos] = palavra;
        pos++;
    }
    return true;
}

// Faz o mapeamento dos dados que estaro inicialmente na
// memria do simulador
bool Risc::MapearMemoria(string arq)
{
    ifstream entrada(arq.c_str());
    string palavra = "";
    int complemento = 0;

    string posicao, valor;
    while (!entrada.eof())
    {
        palavra = "";
        entrada >> posicao;
        entrada >> valor;
        complemento = 16 - valor.length();
        for (int i = 0; i < complemento; i++)
            palavra += '0';

        palavra += valor;
        memoria[ToInt(posicao)] = palavra;
    }
    return true;
}

// Representa um numero passado na sua forma binria
// complemento indica se o sistema deve dar o numero
// no formato 16 bits ou no
string Risc::ToBit(unsigned n, bool complemento)
{
    string valor;
    const int shift = 15;
    const unsigned mascara = 1 << shift;

    for (unsigned i = 1; i <= shift + 1; i++)
    {
        valor += (n & mascara ? '1' : '0');
        n <<= 1;
    }

    bool primeiro = true;
    if (complemento)
    {
        string final = "";
        for (int i = 0; i < valor.length(); i++)
        {
            while(valor[i] == '0' && primeiro)
                i++;
            primeiro = false;
            final += valor[i];
        }
        return final;
    }

    return valor;
}

// Pega um conjunto de bits "0's e 1's dispostos na forma de
// string e os passa para o correspondente decimal
unsigned Risc::ToInt(string s)
{
    double valor = 0;
    int size = s.length() - 1;
    int i = 0, j = 0;

    while (i <= size)
    {
        if (s[size - i] != ' ')
        {
            valor += int(pow(2.0, j) * (s[size - i] - 48));
            j++;
        }
        i++;
    }
    return (unsigned)valor;
}

// Executa todas as intrues passadas ao simulador at que seja
// encontrada a instruo HALT no programa.
// Caso HALT no seja passado o programa entra em loop infinito
void Risc::ProcessarTudo()
{
    emMemoria = false;

    while (cache[0] != HALT)
    {
        // Ciclo 1
        ciclosClock++;
        BuscarInstrucao();

        // Ciclo 2
        ciclosClock++;
        AtualizarPC();

        // Ciclo 3
        ciclosClock++;
        Decodificar();

        // Ciclo 4
        ciclosClock++;
        VerificarOperandos();

        // Ciclo 5
        ciclosClock++;
        if (emMemoria)
            BuscarOperandos();

        // Ciclo 6
        ciclosClock++;
        Executar();

        // Ciclo 7
        ciclosClock++;
        EscreverResultado();
    }
}

// Processa cada ciclo de clock por vez.
// Ser uma funo til quando for desenvolvida a interface
// grfica. Como isso ser possvel disponibilizar um boto
// no qual o usurio poder acompanhar passo-a-passo o que
// seu programa est fazendo
bool Risc::ProcessarPassoAPasso()
{
    int passo = ciclosClock % 7;

    if (cache[0] != HALT)
    {
        switch (passo)
        {
            case 0:
                // Ciclo 1
                emMemoria = false;
                ciclosClock++;
                BuscarInstrucao();
                break;
            case 1:
                // Ciclo 2
                ciclosClock++;
                AtualizarPC();
                break;
            case 2:
                // Ciclo 3
                ciclosClock++;
                Decodificar();
                break;
            case 3:
                // Ciclo 4
                ciclosClock++;
                VerificarOperandos();
                break;
            case 4:
                // Ciclo 5
                ciclosClock++;
                if (emMemoria)
                    BuscarOperandos();
                break;
            case 5:
                // Ciclo 6
                ciclosClock++;
                Executar();
                break;
            case 6:
                // Ciclo 7
                ciclosClock++;
                if (cache[0] != HALT && cache[0] != NOP && cache[0] != LOAD &&
                    cache[0] != STORE)
                    EscreverResultado();
                break;
        }
        return true;
    }
    else
        return false;
}

void Risc::BuscarInstrucao()
{
    int i = ToInt(registradores[0]);
    registradores[1] = memoria[i];
}

void Risc::AtualizarPC()
{
    int pc = ToInt(registradores[0]);
    pc++;
    string bit = ToBit(pc);
    registradores[0] = bit;
}

void Risc::Decodificar()
{
    cache[0] = ToInt(registradores[1].substr(0,4));
    cache[1] = cache[2] = cache[3] = 0;

    switch (cache[0])
    {
        case NOP:
            break;
        case ADD:
        case SUB:
        case MULT:
        case DIV:
        case MOD:
        case AND:
        case OR:
        case XOR:
        case SHIFTL:
        case SHIFTR:
            cache[1] = ToInt(registradores[1].substr(4,4));
            cache[2] = ToInt(registradores[1].substr(8,4));
            cache[3] = ToInt(registradores[1].substr(12,4));
            break;
        case CPOP:
        case NOT:
            cache[1] = ToInt(registradores[1].substr(4,4));
            cache[2] = ToInt(registradores[1].substr(8,4));
            break;
        case LOAD:
        case STORE:
            cache[1] = ToInt(registradores[1].substr(4,4));
            cache[2] = ToInt(registradores[1].substr(8,8));
            break;
        case HALT:
            break;
    }
}

void Risc::VerificarOperandos()
{
    if (cache[0] == LOAD)
        emMemoria = true;
    else
        emMemoria = false;
}

void Risc::BuscarOperandos()
{
    registradores[cache[1]] = memoria[cache[2]];
}

void Risc::Executar()
{
    unsigned result;
    int op1, op2;

    switch (cache[0])
    {
        case NOP:
            break;
        case ADD:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            result = op1 + op2;
            break;
        case SUB:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            result = op1 - op2;
            break;
        case MULT:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            result = op1 * op2;
            break;
        case DIV:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            if (op2 != 0)
                result = op1 / op2;
            else
                result = 0;
            break;
        case SHIFTL:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            result = op1 << op2;
            break;
        case SHIFTR:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            result = op1 >> op2;
        break;
        case AND:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            result = op1 & op2;
            break;
        case OR:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            result = op1 | op2;
            break;
        case XOR:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            result = op1 ^ op2;
            break;
        case NOT:
            op1 = ToInt(registradores[cache[2]]);
            result = ~op1;
            break;
        case CPOP:
            op1 = ToInt(registradores[cache[2]]);
            result = op1;
            break;
        case LOAD:
            break;
        case STORE:
            memoria[cache[2]] = registradores[cache[1]];
            break;
        case MOD:
            op1 = ToInt(registradores[cache[2]]);
            op2 = ToInt(registradores[cache[3]]);
            if (op2 != 0)
                result = op1 % op2;
            else
                result = 0;
            break;
        case HALT:
            break;
    }
    cache[4] = result;
}

void Risc::EscreverResultado()
{
    registradores[cache[1]] = ToBit(cache[4]);
}

void Risc::ImprimirStatus()
{
    cout << "------------------------------------------" << endl;
    cout << "Ciclo de Clock = " << ciclosClock << endl;
    cout << "\nMemoria" << endl;
    for (int i=0; i < 256; i++)
        if (memoria[i] != "")
            cout << "[" << i << "]" << '\t' << memoria[i] << endl;

    cout << "\nRegistradores" << endl;
    for (int i=0; i < 16; i++)
        if (registradores[i] != "")
            cout << "[" << i << "]" << '\t' << registradores[i] << endl;

    cout << "\nCache" << endl;
    for (int i=0; i < 5; i++)
        cout << "[" << i << "]" << '\t' << cache[i] << endl;
}

#endif
